/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::ArticulatedObjectLocalizerDynamicSimulation
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ArticulatedObjectLocalizerDynamicSimulation.h"

// Include headers you only need in function definitions in the .cpp.
#include <algorithm>

#include <Eigen/Core>

#include <ArmarXCore/core/PackagePath.h>
#include <ArmarXCore/core/system/ArmarXDataPath.h>

#include <VirtualRobot/Robot.h>
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/VirtualRobot.h>

#include <RobotAPI/libraries/core/Pose.h>

namespace armarx
{

    armarx::PropertyDefinitionsPtr ArticulatedObjectLocalizerDynamicSimulation::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());

        articulatedObjectWriter->registerPropertyDefinitions(defs);
        articulatedObjectReader->registerPropertyDefinitions(defs);

        defs->component(simulator);

        defs->optional(p.cycleTime, "cycleTime");
        defs->optional(p.objects, "objects", "The objects to (fake) localize.");

        return defs;
    }

    ArticulatedObjectLocalizerDynamicSimulation::ArticulatedObjectLocalizerDynamicSimulation()
        :
        articulatedObjectWriter(new ::armarx::armem::articulated_object::Writer(memoryNameSystem)),
        articulatedObjectReader(new ::armarx::armem::articulated_object::Reader(memoryNameSystem)) {}


    void ArticulatedObjectLocalizerDynamicSimulation::onInitComponent()
    {
    }


    void ArticulatedObjectLocalizerDynamicSimulation::onConnectComponent()
    {
        articulatedObjectWriter->connect();
        articulatedObjectReader->connect();


        setupObjects();


        task = new PeriodicTask<ArticulatedObjectLocalizerDynamicSimulation>(
            this,
            &ArticulatedObjectLocalizerDynamicSimulation::localizeObjects,
            p.cycleTime,
            false,
            "localizeArticulatedObjects");
        task->start();
        task->setDelayWarningTolerance(100);
    }


    void ArticulatedObjectLocalizerDynamicSimulation::onDisconnectComponent()
    {
        task->stop();
    }


    void ArticulatedObjectLocalizerDynamicSimulation::onExitComponent()
    {
    }


    std::string ArticulatedObjectLocalizerDynamicSimulation::getDefaultName() const
    {
        return "ArticulatedObjectLocalizerDynamicSimulation";
    }

    void ArticulatedObjectLocalizerDynamicSimulation::setupObjects()
    {
        const auto descriptions = articulatedObjectReader->queryDescriptions(IceUtil::Time::now());

        const auto tryRegister = [&](const std::string & name)
        {
            ARMARX_INFO << "Checking element '" << name << "' ...";

            const auto isKnownObject = [&name](const auto description) -> bool
            {
                return description.name == name;
            };

            const auto it = std::find_if(descriptions.begin(), descriptions.end(), isKnownObject);
            if (it == descriptions.end())
            {
                ARMARX_INFO << "Skipping element '" << name << "' as it is either a robot or an unknown object";
                return;
            }

            ARMARX_INFO << "Found articulated object '" << name << "'";

            const auto robot = VirtualRobot::RobotIO::loadRobot(ArmarXDataPath::resolvePath(it->xml.serialize().path), VirtualRobot::RobotIO::eStructure);
            if (robot)
            {
                knownObjects[name] = robot;
            }
        };

        // robots can be either active or passive
        const auto robots = simulator->getRobotNames();
        std::for_each(robots.begin(), robots.end(), tryRegister);
    }

    armem::articulated_object::ArticulatedObject convert(const VirtualRobot::Robot& obj, const armem::Time& timestamp)
    {
        ARMARX_DEBUG << "Filename is " << obj.getFilename();

        return
            armem::articulated_object::ArticulatedObject
        {
            .description = {
                .name = obj.getName(),
                .xml = PackagePath(armarx::ArmarXDataPath::getProject({"ArmarXObjects"}, obj.getFilename()), obj.getFilename())
            },
            .instance = "", // TODO(fabian.reister):
            .config = {
                .timestamp = timestamp,
                .globalPose = Eigen::Affine3f(obj.getRootNode()->getGlobalPose()),
                .jointMap = obj.getJointValues()
            },
            .timestamp = timestamp
        };
    }

    void ArticulatedObjectLocalizerDynamicSimulation::localizeObjects()
    {
        // remove objects if the no longer exist in the scene
        // knownObjects.erase(std::remove_if(knownObjects.begin(), knownObjects.end(), [&](const auto&p) -> bool{
        //     const auto& name = p.first;
        //     if(not simulator->hasRobot(name))
        //     {
        //         ARMARX_INFO << "Articulated object '" << name << "' was removed from the scene.";
        //         return true;
        //     }

        //     return false;
        // }), knownObjects.end());

        for (const auto& [name, robot] : knownObjects)
        {
            const PoseBasePtr poseBase = simulator->getRobotPose(name);
            if (not poseBase)
            {
                ARMARX_WARNING << deactivateSpam(10) << "No articulatd object pose from simulator...";
                continue;
            }
            robot->setGlobalPose(PosePtr::dynamicCast(poseBase)->toEigen());

            const auto jointMap = simulator->getRobotJointAngles(name);
            robot->setJointValues(jointMap);

            armarx::armem::articulated_object::ArticulatedObject obj = convert(*robot, IceUtil::Time::now());
            articulatedObjectWriter->store(obj);
        }
    }


}  // namespace armarx
