/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::SimulatorToArviz
 * @author     Peter Reiner ( ufekv at student dot kit dot edu )
 * @date       2020
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "SimulatorToArviz.h"
#include <VirtualRobot/RuntimeEnvironment.h>


namespace armarx
{

    SimulatorToArvizPropertyDefinitions::SimulatorToArvizPropertyDefinitions(std::string prefix) :
        armarx::ComponentPropertyDefinitions(prefix)
    {
        //defineRequiredProperty<std::string>("PropertyName", "Description");
        //defineOptionalProperty<std::string>("PropertyName", "DefaultValue", "Description");

        //defineOptionalProperty<std::string>("DebugObserverName", "DebugObserver",
        //                                  "Name of the topic the DebugObserver listens to.");
        //defineOptionalProperty<std::string>("MyProxyName", "MyProxy", "Name of the proxy to be used");
        defineOptionalProperty<std::string>("PriorKnowledgeName", "PriorKnowledge", "Prior knowledge name");
    }

    SimulatorToArviz::SimulatorToArviz()
    {

    }

    void SimulatorToArviz::reportSceneUpdated(const SceneVisuData& scene, const ::Ice::Current&)
    {
        /*if (!scene)
        {
            ARMARX_ERROR << "No scene available";
            return;
        }*/
        _SceneData = scene;

        armarx::viz::Layer floorLayer = arviz.layer("FloorLayer");
        armarx::viz::Layer robotLayer = arviz.layer("RobotLayer");
        armarx::viz::Layer objectLayer = arviz.layer("ObjectLayer");
        handleRobotLayer(scene.robots, robotLayer);
        handleObjectLayer(scene.objects, objectLayer);

        if (scene.floor)
        {
            ARMARX_DEBUG << "has floor: " << scene.floor << " with filename: " << scene.floorTextureFile;
            std::string filename = "images/Floor20x20.xml";
            VirtualRobot::RuntimeEnvironment::getDataFileAbsolute(filename);
            floorLayer.add(armarx::viz::Object("Floor").file("", filename));
        }
        arviz.enqueueLayer(floorLayer);
        arviz.enqueueLayer(robotLayer);
        arviz.enqueueLayer(objectLayer);
        arviz.commit();
    }

    void SimulatorToArviz::handleRobotLayer(const RobotVisuList& robotList, armarx::viz::Layer& robotLayer)
    {
        for (const auto& e : robotList)
        {
            Eigen::Vector3f position(e.pose->position->x,
                                     e.pose->position->y,
                                     e.pose->position->z);
            Eigen::Quaternionf orientation(e.pose->orientation->qw,
                                           e.pose->orientation->qx,
                                           e.pose->orientation->qy,
                                           e.pose->orientation->qz);
            ARMARX_DEBUG << "Drawing robot " << e.name << " using file: " << e.robotFile << " and joint values " << e.jointValues;
            armarx::viz::Robot robot = armarx::viz::Robot(e.name)
                                       .file(e.project, e.robotFile)
                                       .useFullModel()
                                       .joints(e.jointValues)
                                       .useOriginalColor()
                                       .scale(e.scaling)
                                       .pose(position, orientation);

            robotLayer.add(robot);
        }
    }

    void SimulatorToArviz::handleObjectPrimitiveData(const ObjectVisuData& _object, armarx::viz::Layer& objectLayer)
    {
        const PoseBasePtr& objectPose = _object.objectPoses.at(_object.name);
        Eigen::Vector3f position(objectPose->position->x, objectPose->position->y, objectPose->position->z);
        Eigen::Quaternionf orientation(objectPose->orientation->qw, objectPose->orientation->qx, objectPose->orientation->qy, objectPose->orientation->qz);

        switch (_object.objectPrimitiveData->type)
        {
            case ObjectType::Box:
            {
                BoxVisuPrimitivePtr boxObject = BoxVisuPrimitivePtr::dynamicCast(_object.objectPrimitiveData);
                //armarx::viz::Box box = armarx::viz::Box(_object.name)
                objectLayer.add(armarx::viz::Box(_object.name)
                                .size(Eigen::Vector3f{boxObject->width, boxObject->height, boxObject->depth})
                                .color(boxObject->color.r, boxObject->color.g, boxObject->color.b)
                                .position(position)
                                .orientation(orientation));
            }
            break;
            case ObjectType::Sphere:
            {
                SphereVisuPrimitivePtr sphereObject = SphereVisuPrimitivePtr::dynamicCast(_object.objectPrimitiveData);
                //armarx::viz::Sphere sphere = armarx::viz::Sphere(_object.name)
                objectLayer.add(armarx::viz::Sphere(_object.name)
                                .radius(sphereObject->radius)
                                .color(sphereObject->color.r, sphereObject->color.g, sphereObject->color.b)
                                .position(position)
                                .orientation(orientation));
            }
            break;
            case ObjectType::Cylinder:
            {
                CylinderVisuPrimitivePtr cylinderObject = CylinderVisuPrimitivePtr::dynamicCast(_object.objectPrimitiveData);
                //armarx::viz::Cylinder cylinder = armarx::viz::Cylinder(_object.name)
                objectLayer.add(armarx::viz::Cylinder(_object.name)
                                .radius(cylinderObject->radius)
                                .height(cylinderObject->length)
                                .color(cylinderObject->color.r, cylinderObject->color.g, cylinderObject->color.b)
                                .position(position)
                                .orientation(orientation));
            }
            break;
        }
    }

    void SimulatorToArviz::handleObjectFromFile(const ObjectVisuData& _object, armarx::viz::Layer& objectLayer)
    {
        const PoseBasePtr& objectPose = _object.objectPoses.at(_object.name);

        Eigen::Vector3f position(objectPose->position->x,
                                 objectPose->position->y,
                                 objectPose->position->z);
        Eigen::Quaternionf orientation(objectPose->orientation->qw,
                                       objectPose->orientation->qx,
                                       objectPose->orientation->qy,
                                       objectPose->orientation->qz);

        objectLayer.add(armarx::viz::Object(_object.name)
                        .file(_object.project, _object.filename)
                        .position(position)
                        .orientation(orientation));
    }

    void SimulatorToArviz::handleObjectFromObjectClass(const ObjectVisuData& _object, armarx::viz::Layer& objectLayer)
    {

        const PoseBasePtr& objectPose = _object.objectPoses.at(_object.name);

        Eigen::Vector3f position(objectPose->position->x,
                                 objectPose->position->y,
                                 objectPose->position->z);
        Eigen::Quaternionf orientation(objectPose->orientation->qw,
                                       objectPose->orientation->qx,
                                       objectPose->orientation->qy,
                                       objectPose->orientation->qz);
        std::string filename = getClassFilename(_object.objectClassName);
        std::string projectname = "";
        objectLayer.add(armarx::viz::Object(_object.name)
                        .file(projectname, filename)
                        .position(position)
                        .orientation(orientation));

    }

    std::string SimulatorToArviz::getClassFilename(const std::string& className)
    {
        auto it = _classFilenameCache.find(className);
        if (it != _classFilenameCache.end())
        {
            return it->second;
        }
        else
        {
            std::optional<memoryx::ObjectClassWrapper> objectClass = _objectClassSegment.getClass(className);
            if (!objectClass)
            {
                return "";
            }
            memoryx::EntityWrappers::SimoxObjectWrapperPtr wrapper =
                objectClass->classInMemory->getWrapper<memoryx::EntityWrappers::SimoxObjectWrapper>();
            std::string filename = wrapper->getManipulationObjectFileName();

            _classFilenameCache[className] = filename;

            return filename;
        }
    }

    void SimulatorToArviz::handleObjectLayer(const ObjectVisuList& objectList, armarx::viz::Layer& objectLayer)
    {
        for (const auto& e : objectList)
        {
            if (e.objectPrimitiveData)
            {
                handleObjectPrimitiveData(e, objectLayer);
            }
            else if (!e.filename.empty())
            {
                handleObjectFromFile(e, objectLayer);
            }
            else
            {
                handleObjectFromObjectClass(e, objectLayer);
            }
        }
    }

    std::string SimulatorToArviz::getDefaultName() const
    {
        return "SimulatorToArviz";
    }


    void SimulatorToArviz::onInitComponent()
    {
        // Register offered topices and used proxies here.
        //offeringTopicFromProperty("DebugObserverName");
        // debugDrawer.offeringTopic(*this);  // Calls this->offeringTopic().

        // Using a proxy will cause the component to wait until the proxy is available.
        // usingProxyFromProperty("MyProxyName");
        usingTopic("SimulatorListenerInterface");
        usingTopic("SimulatorVisuUpdates");
        usingProxyFromProperty("PriorKnowledgeName");
    }


    void SimulatorToArviz::onConnectComponent()
    {
        // Get topics and proxies here. Pass the *InterfacePrx type as template argument.
        //getTopicFromProperty(debugObserver, "DebugObserverName");
        // debugDrawer.getTopic(*this);  // Calls this->getTopic().
        //getTopic("ReportVisuTopicName");
        // getProxyFromProperty<MyProxyInterfacePrx>("MyProxyName");
        getProxyFromProperty(_priorKnowledge, "PriorKnowledgeName");
        _objectClassSegment.initFromProxy(_priorKnowledge);
    }


    void SimulatorToArviz::onDisconnectComponent()
    {

    }


    void SimulatorToArviz::onExitComponent()
    {

    }




    armarx::PropertyDefinitionsPtr SimulatorToArviz::createPropertyDefinitions()
    {
        return armarx::PropertyDefinitionsPtr(new SimulatorToArvizPropertyDefinitions(
                getConfigIdentifier()));
    }
}
