/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ArmarXSimulation::ArmarXObjects::ClutteredSceneGenerator
 * @author     Patrick Hegemann ( patrick dot hegemann at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "ClutteredSceneGenerator.h"

#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>

#include <SimoxUtility/math/pose/pose.h>

#include <VirtualRobot/CollisionDetection/CollisionChecker.h>
#include <VirtualRobot/Obstacle.h>
#include <VirtualRobot/SceneObjectSet.h>
#include <VirtualRobot/XML/RobotIO.h>

namespace armarx
{
    ClutteredSceneGenerator::ClutteredSceneGenerator(const armarx::SimulatorInterfacePrx& simulator,
                                                     const ClutteredSceneGenerator::Config &config)
    {
        this->simulator = simulator;
        this->config = config;

        createCollisionWalls();
    }

    ClutteredSceneGenerator::~ClutteredSceneGenerator() {
        deleteLocalObjectCopies();
        // TODO: Clean up the collision walls
    }

    void ClutteredSceneGenerator::generateScene(int seed) {
        simulator->stop();
        deleteLocalObjectCopies();

        simulator->reInitialize();
        rnd.seed(seed);
        int objectCount = 0;
        int objectsPerSet = 0;

        while (objectCount < config.amountObjects
               && (objectsPerSet < config.objectsPerSetLimit || config.objectsPerSetLimit == 0)
               && !config.objectSetFiles.empty())
        {
            // Spawn one object of each set
            for (auto & currentSetObjectFiles : config.objectSetFiles)
            {
                // Select random object from this set
                boost::uniform_int<> uniform(0, currentSetObjectFiles.size() - 1);
                int objectNr = uniform(rnd);
                string filename = currentSetObjectFiles[objectNr];

                // Spawn this object in the simulator
                string robotName = simulator->addRobot(filename);
                simulator->setRobotPose(robotName, randomObjectPose(objectCount));

                // Create local copy of the object
                makeLocalObjectCopy(robotName, filename);

                // Stop generating objects once we have enough
                objectCount++;
                if (objectCount >= config.amountObjects)
                {
                    break;
                }
            }

            objectsPerSet++;
        }

        // Let objects fall onto the table
        dropObjects();

        // Try to reset some objects until all of them are on the table
        if (config.autoResetInvalidObjects)
        {
            resetInvalidObjects();
        }

        simulator->start();
    }

    void ClutteredSceneGenerator::createCollisionWalls() {
        // Initialize invisible walls for object collision checking
        const float wallHeight = 1000;
        frontWall = Obstacle::createBox(config.boxW, 1, wallHeight);
        backWall = Obstacle::createBox(config.boxW, 1, wallHeight);
        leftWall = Obstacle::createBox(1, config.boxH, wallHeight);
        rightWall = Obstacle::createBox(1, config.boxH, wallHeight);

        const float wallZ = config.boxZ + config.boxD + wallHeight / 2;
        frontWall->setGlobalPose(simox::math::pose(Eigen::Vector3f{config.boxX, config.boxY - config.boxH / 2.0f, wallZ}));
        backWall->setGlobalPose(simox::math::pose(Eigen::Vector3f{config.boxX, config.boxY + config.boxH / 2.0f, wallZ}));
        leftWall->setGlobalPose(simox::math::pose(Eigen::Vector3f{config.boxX - config.boxW / 2.0f, config.boxY, wallZ}));
        rightWall->setGlobalPose(simox::math::pose(Eigen::Vector3f{config.boxX + config.boxW / 2.0f, config.boxY, wallZ}));

        walls.reset(new SceneObjectSet("walls"));
        walls->addSceneObject(frontWall);
        walls->addSceneObject(backWall);
        walls->addSceneObject(leftWall);
        walls->addSceneObject(rightWall);
    }

    PosePtr ClutteredSceneGenerator::randomObjectPose(int heightIndex) {
        // Object position and angle random distributions
        boost::uniform_real<> objectPositionDistX(config.minObjectX, config.maxObjectX);
        boost::uniform_real<> objectPositionDistY(config.minObjectY, config.maxObjectY);
        boost::uniform_real<> objectAngleDist(0, 2 * M_PI);

        // Sample random position and rotation around z-axis
        float x = objectPositionDistX(rnd);
        float y = objectPositionDistY(rnd);
        float z = config.minObjectZ + (float)heightIndex * config.objectSpacingZ;
        float angle = objectAngleDist(rnd);

        Eigen::AngleAxisf rotation(angle, Eigen::Vector3f::UnitZ());
        PosePtr pose(new Pose(simox::math::pose(Eigen::Vector3f{x, y, z}, rotation)));
        return pose;
    }

    void ClutteredSceneGenerator::deleteLocalObjectCopies() {
        // TODO: Do we have to clean up the robot instances individually?
        localObjectCopies.clear();
    }

    void ClutteredSceneGenerator::makeLocalObjectCopy(const string& objectName, const string& filename) {
        // Load robot from file and get pose from simulator
        RobotPtr robot = RobotIO::loadRobot(filename);
        robot->setName(objectName);
        PoseBasePtr poseBase = simulator->getRobotPose(objectName);
        PosePtr pose(new Pose(poseBase->position, poseBase->orientation));
        robot->setGlobalPose(pose->toEigen());

        localObjectCopies.push_back(robot);
    }

    void ClutteredSceneGenerator::updateLocalObjectCopies() {
        // Retrieve new poses from simulator
        for (const RobotPtr& robot : localObjectCopies)
        {
            PoseBasePtr poseBase = simulator->getRobotPose(robot->getName());
            PosePtr pose(new Pose(poseBase->position, poseBase->orientation));
            robot->setGlobalPose(pose->toEigen());
        }
    }

    void ClutteredSceneGenerator::dropObjects() {
        simulator->stop();
        for (int t = 0; t < config.fallingSteps; ++t)
        {
            simulator->step();
        }
        simulator->start();
    }

    void ClutteredSceneGenerator::resetInvalidObjects() {
        CollisionCheckerPtr col = CollisionChecker::getGlobalCollisionChecker();
        bool anyRespawn;

        // Try to reset the objects a number of times, or until all objects are inside the box
        for (int i = 0; i < config.maxResetTries; ++i)
        {
            updateLocalObjectCopies();

            // TODO: get rid of magic number
            int heightIndex = 6;
            anyRespawn = false;

            // Check and reset every object
            for (const RobotPtr& robot : localObjectCopies)
            {
                // Check if object collides with the imaginary walls around the box
                bool collision = false;
                for (const auto& colModel : robot->getCollisionModels())
                {
                    collision |= col->checkCollision(colModel, walls);
                }
                Eigen::Vector3f objectPosition = robot->getGlobalPosition();

                // Check if objects is outside of table
                // TODO: Magic number
                bool outsideOfTable = objectPosition[2] < config.minObjectZOtherwiseReset;

                // Respawn object if necessary
                if (collision || outsideOfTable)
                {
                    anyRespawn = true;
                    PosePtr newPose = randomObjectPose(heightIndex);
                    simulator->setRobotPose(robot->getName(), newPose);
                    heightIndex++;
                }
            }

            // Let objects fall again if anything has changed
            if (!anyRespawn)
            {
                break;
            }
            dropObjects();
        }
    }

    const ClutteredSceneGenerator::Config &ClutteredSceneGenerator::getConfig() const {
        return config;
    }

    void ClutteredSceneGenerator::setConfig(const ClutteredSceneGenerator::Config& conf) {
        ClutteredSceneGenerator::config = conf;
    }

    const SimulatorInterfacePrx &ClutteredSceneGenerator::getSimulator() const {
        return simulator;
    }

    void ClutteredSceneGenerator::setSimulator(const SimulatorInterfacePrx &sim) {
        ClutteredSceneGenerator::simulator = sim;
    }
}
